/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cmd.c -- Quake script command processing module

#include "quakedef.h"

void Cmd_ForwardToServer (void);

#define    MAX_ALIAS_NAME    32

struct cmdalias_t
{
    std::string name, value;
};

std::vector<cmdalias_t> cmd_alias;
qboolean cmd_wait;

//=============================================================================

/*
============
Cmd_Wait_f

Causes execution of the remainder of the command buffer to be delayed until
next frame.  This allows commands like:
bind g "impulse 5 ; +attack ; wait ; -attack ; impulse 2"
============
*/
void Cmd_Wait_f (void)
{
    cmd_wait = true;
}

/*
=============================================================================

                        COMMAND BUFFER

=============================================================================
*/

sizebuf_t    cmd_text;

/*
============
Cbuf_Init
============
*/
void Cbuf_Init (void)
{
    SZ_Alloc (&cmd_text, 8192);        // space for commands and script files
}


/*
============
Cbuf_AddText

Adds command text at the end of the buffer
============
*/
void Cbuf_AddText (const std::string &text)
{
    size_t l = text.length();

    if (cmd_text.cursize + (int)l >= cmd_text.maxsize)
    {
        Con_Printf ("Cbuf_AddText: overflow\n");
        return;
    }

    SZ_Write (&cmd_text, text.c_str(), l);
}


/*
============
Cbuf_InsertText

Adds command text immediately after the current command
Adds a \n to the text
FIXME: actually change the command buffer to do less copying
============
*/
void Cbuf_InsertText (const std::string &text)
{
    std::vector<uint8_t> temp;

    // copy off any commands still remaining in the exec buffer
    if (cmd_text.cursize)
    {
        temp.resize(cmd_text.cursize);
        memcpy(&temp[0], cmd_text.data, cmd_text.cursize);
        SZ_Clear(&cmd_text);
    }

    // add the entire text of the file
    Cbuf_AddText (text);

    // add the copied off data
    if (temp.size())
        SZ_Write(&cmd_text, &temp[0], temp.size());
}

/*
============
Cbuf_Execute
============
*/
void Cbuf_Execute (void)
{
    int        i;
    char    *text;
    char    line[1024];
    int        quotes;

    while (cmd_text.cursize)
    {
// find a \n or ; line break
        text = (char *)cmd_text.data;

        quotes = 0;
        for (i=0 ; i< cmd_text.cursize ; i++)
        {
            if (text[i] == '"')
                quotes++;
            if ( !(quotes&1) &&  text[i] == ';')
                break;    // don't break if inside a quoted string
            if (text[i] == '\n')
                break;
        }


        memcpy (line, text, i);
        line[i] = 0;

// delete the text from the command buffer and move remaining commands down
// this is necessary because commands (exec, alias) can insert data at the
// beginning of the text buffer

        if (i == cmd_text.cursize)
            cmd_text.cursize = 0;
        else
        {
            i++;
            cmd_text.cursize -= i;
            memcpy (text, text+i, cmd_text.cursize);
        }

// execute the command line
        Cmd_ExecuteString (line, src_command);

        if (cmd_wait)
        {    // skip out while text still remains in buffer, leaving it
            // for next frame
            cmd_wait = false;
            break;
        }
    }
}

/*
==============================================================================

                        SCRIPT COMMANDS

==============================================================================
*/

/*
===============
Cmd_Exec_f
===============
*/
void Cmd_Exec_f (void)
{
    if (Cmd_Argc () != 2)
    {
        Con_Printf ("exec <filename> : execute a script file\n");
        return;
    }
    
    File fp;
    if (!fp.OpenRead(Cmd_Argv(1)))
    {
        Con_Printf("couldn't exec %s, cannot open file\n", Cmd_Argv(1).c_str());
        return;
    }

    char *f = (char *)fp.Read();
    if (!f)
    {
        Con_Printf ("couldn't exec %s, cannot read file\n", Cmd_Argv(1).c_str());
        return;
    }
    Con_Printf ("execing %s\n", Cmd_Argv(1).c_str());
    
    Cbuf_InsertText (f);
    
    delete[] f;
    fp.Close();
}


/*
===============
Cmd_Echo_f

Just prints the rest of the line to the console
===============
*/
void Cmd_Echo_f (void)
{
    for (size_t i = 1 ; i < Cmd_Argc() ; i++)
        Con_Printf("%s ", Cmd_Argv(i).c_str());
    Con_Printf("\n");
}

/*
===============
Cmd_Alias_f

Creates a new command that executes a command string (possibly ; seperated)
===============
*/
void Cmd_Alias_f (void)
{
    cmdalias_t a;
    std::vector<cmdalias_t>::iterator iter;
    int i, c;
    std::string name;

    if (Cmd_Argc() == 1)
    {
        Con_Printf ("Current alias commands:\n");
        for (iter = cmd_alias.begin() ; iter != cmd_alias.end() ; ++iter)
            Con_Printf ("%s : %s\n", (*iter).name.c_str(), (*iter).value.c_str());
        return;
    }

    name = Cmd_Argv(1);
    boost::to_lower(name);

    // if the alias already exists, erase it
    for (iter = cmd_alias.begin() ; iter != cmd_alias.end() ; ++iter)
    {
        if ((*iter).name == name)
        {
            cmd_alias.erase(iter);
            break;
        }
    }

    a.name = name;
    a.value.erase();

// copy the rest of the command line
    c = Cmd_Argc();
    for (i = 2 ; i < c ; i++)
    {
        a.value += Cmd_Argv(i);
        if (i != c)
            a.value += " ";
    }
    a.value += "\n";
    
    cmd_alias.push_back(a);
}

/*
=============================================================================

                    COMMAND EXECUTION

=============================================================================
*/

typedef struct cmd_function_s
{
    struct cmd_function_s    *next;
    char                    *name;
    xcommand_t                function;
} cmd_function_t;


#define    MAX_ARGS        80

static std::vector<std::string> cmd_argv;
static std::string cmd_args;

cmd_source_t    cmd_source;


static    cmd_function_t    *cmd_functions;        // possible commands to execute

/*
============
Cmd_Init
============
*/
void Cmd_Init (void)
{
//
// register our commands
//
    Cmd_AddCommand ("exec",Cmd_Exec_f);
    Cmd_AddCommand ("echo",Cmd_Echo_f);
    Cmd_AddCommand ("alias",Cmd_Alias_f);
    Cmd_AddCommand ("cmd", Cmd_ForwardToServer);
    Cmd_AddCommand ("wait", Cmd_Wait_f);
}

/*
============
Cmd_Argc
============
*/
size_t Cmd_Argc()
{
    return cmd_argv.size();
}

/*
============
Cmd_Argv
============
*/
std::string Cmd_Argv(size_t arg)
{
    if (arg < 0 || arg >= cmd_argv.size())
        return std::string();
    return cmd_argv[arg];
}

/*
============
Cmd_Args
============
*/
std::string Cmd_Args()
{
    return cmd_args;
}


/*
============
Cmd_TokenizeString

Parses the given string into command line tokens.
============
*/
void Cmd_TokenizeString (char *text)
{
    // clear the args from the last string
    cmd_argv.clear();
    cmd_args.erase();

    while (1)
    {
        // skip whitespace up to a /n
        while (*text && *text <= ' ' && *text != '\n')
            text++;

        if (*text == '\n')
        {
            // a newline seperates commands in the buffer
            text++;
            break;
        }

        if (!*text)
            return;

        if (cmd_argv.size() == 1)
            cmd_args = text;

        text = COM_Parse (text);
        if (!text)
            return;

        std::string str(com_token);
        cmd_argv.push_back(str);
    }
}


/*
============
Cmd_AddCommand
============
*/
void    Cmd_AddCommand (char *cmd_name, xcommand_t function)
{
    cmd_function_t    *cmd;

    if (host_initialized)    // because hunk allocation would get stomped
        Sys_Error ("Cmd_AddCommand after host_initialized");

// fail if the command is a variable name
    if (Cvar_VariableString(cmd_name)[0])
    {
        Con_Printf ("Cmd_AddCommand: %s already defined as a var\n", cmd_name);
        return;
    }

// fail if the command already exists
    for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
    {
        if (!strcmp (cmd_name, cmd->name))
        {
            Con_Printf ("Cmd_AddCommand: %s already defined\n", cmd_name);
            return;
        }
    }

    cmd = (cmd_function_t *)Hunk_Alloc (sizeof(cmd_function_t));
    cmd->name = cmd_name;
    cmd->function = function;
    cmd->next = cmd_functions;
    cmd_functions = cmd;
}

/*
============
Cmd_Exists
============
*/
qboolean Cmd_Exists(const std::string &cmd_name)
{
    cmd_function_t    *cmd;

    for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
    {
        if (!strcmp (cmd_name.c_str(),cmd->name))
            return true;
    }

    return false;
}



/*
============
Cmd_CompleteCommand
============
*/
char *Cmd_CompleteCommand (char *partial)
{
    cmd_function_t    *cmd;
    int                len;

    len = strlen(partial);

    if (!len)
        return NULL;

// check functions
    for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
        if (!strncmp (partial,cmd->name, len))
            return cmd->name;

    return NULL;
}

/*
============
Cmd_ExecuteString

A complete command line has been parsed, so try to execute it
FIXME: lookupnoadd the token to speed search?
============
*/
void    Cmd_ExecuteString (char *text, cmd_source_t src)
{
    cmd_function_t *cmd;

    cmd_source = src;
    Cmd_TokenizeString (text);

// execute the command line
    if (!Cmd_Argc())
        return;        // no tokens
    
    std::string lower = cmd_argv[0];
    boost::to_lower(lower);

// check functions
    for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
    {
        if (boost::iequals(cmd->name, cmd_argv[0]))
        {
            cmd->function ();
            return;
        }
    }

// check alias
    for (std::vector<cmdalias_t>::iterator iter = cmd_alias.begin() ; 
         iter != cmd_alias.end() ; ++iter)
    {
        if ((*iter).name == lower)
        {
            Cbuf_InsertText ((*iter).value);
            return;
        }
    }

    // check cvars
    if (Cvar_Command())
        return;

    Con_Printf ("Unknown command \"%s\"\n", Cmd_Argv(0).c_str());
}


/*
===================
Cmd_ForwardToServer

Sends the entire command line over to the server
===================
*/
void Cmd_ForwardToServer (void)
{
    if (cls.state != ca_connected)
    {
        Con_Printf ("Can't \"%s\", not connected\n", Cmd_Argv(0).c_str());
        return;
    }

    if (cls.demoplayback)
        return;        // not really connected

    MSG_WriteByte (&cls.message, clc_stringcmd);
    if (!boost::iequals(Cmd_Argv(0), "cmd"))
    {
        SZ_Print (&cls.message, Cmd_Argv(0));
        SZ_Print (&cls.message, " ");
    }
    if (Cmd_Argc() > 1)
        SZ_Print (&cls.message, Cmd_Args());
    else
        SZ_Print (&cls.message, "\n");
}

