/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cvar.c -- dynamic variable tracking

#include "quakedef.h"

class cvar_t
{
public:
    cvar_t() : have_float(false) { }

    std::string name;
    std::string string;

    bool archive;

    float to_f()
    {
        if (!have_float)
        {
            f = Util::lexical_cast<float>(string);
            have_float = true;
        }
        
        return f;
    }

private:
    bool have_float;
    float f;
};

std::vector<cvar_t> cvar_vars;

/*
============
Cvar_FindVar
============
*/
cvar_t *Cvar_FindVar (const std::string &var_name)
{
    std::vector<cvar_t>::iterator iter;
    iter = std::find_if(cvar_vars.begin(), cvar_vars.end(),
        bind(&cvar_t::name, _1) == var_name);
    
    if (iter == cvar_vars.end())
        return NULL;
    return &(*iter);
}

/*
============
Cvar_VariableValue
============
*/
float Cvar_VariableValue (const std::string &var_name)
{
    cvar_t    *var;

    var = Cvar_FindVar(var_name);
    if (!var)
        return 0;
    return var->to_f();
}


/*
============
Cvar_VariableString
============
*/
std::string Cvar_VariableString (const std::string &var_name)
{
    cvar_t *var;

    var = Cvar_FindVar(var_name);
    if (!var)
        return std::string();
    return var->string;
}


/*
============
Cvar_CompleteVariable
============
*/
std::string Cvar_CompleteVariable(const std::string &partial)
{
    if (!partial.length())
        return std::string();

    // check functions
    for (std::vector<cvar_t>::iterator iter = cvar_vars.begin() ;
         iter != cvar_vars.end() ; ++iter)
    {
        if (!(*iter).name.compare(0, partial.length(), partial))
            return (*iter).name;
    }

    return std::string();
}


/*
============
Cvar_Set
============
*/
void Cvar_Set(const std::string &var_name, const std::string &value)
{
    cvar_t *var = Cvar_FindVar (var_name);
    if (!var)
    {
        // there is an error in C code if this happens
        Con_Printf("Cvar_Set: variable %s not found\n", var_name.c_str());
        return;
    }

    var->string = value;
}

/*
============
Cvar_SetValue
============
*/
void Cvar_SetValue(const std::string &var_name, float value)
{
    Cvar_Set(var_name, Util::lexical_cast<std::string>(value));
}


/*
============
Cvar_RegisterVariable

Adds a freestanding variable to the variable list.
============
*/
void Cvar_RegisterVariable(const std::string &name, const std::string &value, qboolean archive)
{
    // first check to see if it has already been defined
    if (Cvar_FindVar(name))
    {
        Con_Printf("Can't register variable %s, already defined\n", name.c_str());
        return;
    }

    // check for overlap with a command
    if (Cmd_Exists(name))
    {
        Con_Printf("Cvar_RegisterVariable: %s is a command\n", name.c_str());
        return;
    }

    cvar_t variable;
    variable.name = name;
    variable.string = value;
    variable.archive = archive;
    cvar_vars.push_back(variable);
}

/*
============
Cvar_Command

Handles variable inspection and changing from the console
============
*/
qboolean Cvar_Command (void)
{
    cvar_t *v;

    // check variables
    v = Cvar_FindVar(Cmd_Argv(0));
    if (!v)
        return false;

    // perform a variable print or set
    if (Cmd_Argc() == 1)
    {
        Con_Printf ("\"%s\" is \"%s\"\n", v->name.c_str(), v->string.c_str());
        return true;
    }

    Cvar_Set (v->name, Cmd_Argv(1).c_str());
    return true;
}


/*
============
Cvar_WriteVariables

Writes lines containing "set variable value" for all variables
with the archive flag set to true.
============
*/
void Cvar_WriteVariables (File &fp)
{
    for (std::vector<cvar_t>::iterator iter = cvar_vars.begin() ;
         iter != cvar_vars.end() ; ++iter)
    {
        if ((*iter).archive)
            fp.Printf("%s \"%s\"\n", (*iter).name.c_str(), (*iter).string.c_str());
    }
}

