/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// snd_dma.c -- main control for any streaming sound output device
#include "quakedef.h"
#include "filesystem/filesystem.h"
#include "snd_internal.h"
#include <SDL/SDL.h>

void S_Play(void);
void S_PlayVol(void);

extern void S_InitPaintBuffer(size_t len);
extern void S_ShutdownPaintBuffer(void);
extern void S_MixSDLAudio (void *unused, unsigned char *stream, int len);

// =======================================================================
// Internal sound data & structures
// =======================================================================

channel_t   channels[MAX_CHANNELS];
int            total_channels;

vec3_t        listener_origin;
vec3_t        listener_forward;
vec3_t        listener_right;
vec3_t        listener_up;
vec_t        sound_nominal_clip_dist=1000.0;


#define    MAX_SFX        512
sfx_t        *known_sfx;        // hunk allocated [MAX_SFX]
int            num_sfx;

sfx_t        *ambient_sfx[NUM_AMBIENTS];

qboolean sound_started = false;
int snd_freq;


/*
================
S_Startup
================
*/
void S_Startup (void)
{
    SDL_AudioSpec fmt;

    sound_started = false;

    fmt.freq = 11025;
    fmt.format = AUDIO_S16SYS;
    fmt.channels = 2;
    fmt.samples = 1024;
    fmt.callback = S_MixSDLAudio;
    fmt.userdata = NULL;

    if (SDL_OpenAudio(&fmt, NULL) < 0)
    {
        Con_SafePrintf ("Couldn't open SDL audio stream\n");
        return;
    }

    S_InitPaintBuffer(fmt.size / 4);
    snd_freq = fmt.freq;

    SDL_PauseAudio(0);
    sound_started = true;
}


/*
================
S_Init
================
*/
void S_Init (void)
{
    Con_Printf("\nSound Initialization\n");

    Cmd_AddCommand("play", S_Play);
    Cmd_AddCommand("playvol", S_PlayVol);
    Cmd_AddCommand("stopsound", S_StopAllSounds);

    Cvar_RegisterVariable("nosound", "0");
    Cvar_RegisterVariable("volume", "0.7", true);
    Cvar_RegisterVariable("bgmvolume", "1", true);
    Cvar_RegisterVariable("ambient_level", "0.3");
    Cvar_RegisterVariable("ambient_fade", "100");

    known_sfx = new sfx_t[MAX_SFX];
    num_sfx = 0;

    S_Startup();

    ambient_sfx[AMBIENT_WATER] = known_sfx + S_PrecacheSound("ambience/water1.wav");
    ambient_sfx[AMBIENT_SKY] = known_sfx + S_PrecacheSound("ambience/wind2.wav");
}


// =======================================================================
// Shutdown sound engine
// =======================================================================

void S_Shutdown(void)
{
    if (!sound_started)
        return;

    sound_started = 0;

    S_ShutdownPaintBuffer();
    SDL_CloseAudio();

    delete[] known_sfx;
}


// =======================================================================
// Load a sound
// =======================================================================

/*
==================
S_FindName
==================
*/
sfx_t *S_FindName(const std::string &name)
{
    int        i;
    sfx_t    *sfx;

    if (!name.length())
        Sys_Error ("S_FindName: NULL\n");

    // see if already loaded
    for (i = 0 ; i < num_sfx ; i++)
    {
        if (known_sfx[i].name == name)
            return &known_sfx[i];
    }

    if (num_sfx == MAX_SFX)
        Sys_Error ("S_FindName: out of sfx_t");

    SDL_LockAudio();

    sfx = &known_sfx[i];
    sfx->name = name;

    num_sfx++;

    SDL_UnlockAudio();
    return sfx;
}


/*
==================
S_PrecacheSound
==================
*/
int S_PrecacheSound(const std::string &name)
{
    sfx_t *sfx = S_FindName (name);

    // cache it in
    SDL_LockAudio();
    S_LoadSound(sfx);
    SDL_UnlockAudio();

    return sfx - known_sfx;
}


//=============================================================================

/*
=================
S_PickChannel
=================
*/
channel_t *S_PickChannel(int entnum, int entchannel)
{
    int ch_idx;
    int first_to_die;
    int highest_pos;

    // check for replacement sound, or find the best one to replace
    first_to_die = -1;
    highest_pos = 0;
    for (ch_idx=NUM_AMBIENTS ; ch_idx < NUM_AMBIENTS + MAX_DYNAMIC_CHANNELS ; ch_idx++)
    {
        if (entchannel != 0    &&    // channel 0 never overrides
            channels[ch_idx].entnum == entnum &&
            (channels[ch_idx].entchannel == entchannel || entchannel == -1) )
        {
            // always override sound from same entity
            first_to_die = ch_idx;
            break;
        }

        // don't let monster sounds override player sounds
        if (channels[ch_idx].entnum == cl.viewentity && entnum != cl.viewentity && channels[ch_idx].sfx)
            continue;

        // always override if not playing
        if (!channels[ch_idx].sfx)
        {
            first_to_die = ch_idx;
            break;
        }

        if (channels[ch_idx].pos > highest_pos)
        {
            highest_pos = channels[ch_idx].pos;
            first_to_die = ch_idx;
        }
    }

    if (first_to_die == -1)
        return NULL;

    if (channels[first_to_die].sfx)
    {
        channels[first_to_die].sfx = NULL;
        channels[first_to_die].pos = 999999;
    }

    return &channels[first_to_die];
}

/*
=================
S_Spatialize
=================
*/
void S_Spatialize(channel_t *ch)
{
    vec_t dot;
    vec_t dist;
    vec_t lscale, rscale, scale;
    vec3_t source_vec;
    sfx_t *snd;

    // anything coming from the view entity will always be full volume
    if (ch->entnum == cl.viewentity)
    {
        ch->leftvol = ch->master_vol;
        ch->rightvol = ch->master_vol;
        return;
    }

    // calculate stereo separation and distance attenuation
    snd = ch->sfx;
    VectorSubtract(ch->origin, listener_origin, source_vec);

    dist = VectorNormalize(source_vec) * ch->dist_mult;

    dot = DotProduct(listener_right, source_vec);

    rscale = 1.0 + dot;
    lscale = 1.0 - dot;

    // add in distance effect
    scale = (1.0 - dist) * rscale;
    ch->rightvol = (int) (ch->master_vol * scale);
    if (ch->rightvol < 0)
        ch->rightvol = 0;

    scale = (1.0 - dist) * lscale;
    ch->leftvol = (int) (ch->master_vol * scale);
    if (ch->leftvol < 0)
        ch->leftvol = 0;
}


// =======================================================================
// Start a sound effect
// =======================================================================

void S_StartSound(int entnum, int entchannel, int soundindex, vec3_t origin, float fvol, float attenuation)
{
    sfx_t *sfx;
    channel_t *target_chan, *check;
    sfxcache_t    *sc;
    int        vol;
    int        ch_idx;
    int        skip;

    if (!sound_started)
        return;
    if (Cvar_VariableValue("nosound"))
        return;
    if (soundindex < 0 || soundindex >= num_sfx)
        return;

    sfx = known_sfx + soundindex;
    if (!sfx)
        return;

    vol = fvol*255;

    SDL_LockAudio();

    // pick a channel to play on
    target_chan = S_PickChannel(entnum, entchannel);
    if (!target_chan)
    {
        SDL_UnlockAudio();
        return;
    }

    // spatialize
    memset (target_chan, 0, sizeof(*target_chan));
    VectorCopy(origin, target_chan->origin);
    target_chan->dist_mult = attenuation / sound_nominal_clip_dist;
    target_chan->master_vol = vol;
    target_chan->entnum = entnum;
    target_chan->entchannel = entchannel;
    S_Spatialize(target_chan);

    if (!target_chan->leftvol && !target_chan->rightvol)
    {
        SDL_UnlockAudio();
        return;     // not audible at all
    }

    // new channel
    sc = S_LoadSound (sfx);
    if (!sc)
    {
        target_chan->sfx = NULL;
        target_chan->pos = 999999;
        SDL_UnlockAudio();
        return;        // couldn't load the sound's data
    }

    if (sc->looppos == -1)
        target_chan->looping = 0;
    else
        target_chan->looping = 1;

    target_chan->sfx = sfx;
    target_chan->pos = 0;

    // if an identical sound has also been started this frame, offset the pos
    // a bit to keep it from just making the first one louder
    check = &channels[NUM_AMBIENTS];
    for (ch_idx=NUM_AMBIENTS ; ch_idx < NUM_AMBIENTS + MAX_DYNAMIC_CHANNELS ; ch_idx++, check++)
    {
        if (check == target_chan)
            continue;
        if (check->sfx == sfx && !check->pos)
        {
            skip = rand () % (int)(0.1 * snd_freq);
            if (skip >= sc->length)
                skip = sc->length - 1;
            target_chan->pos += skip;
            break;
        }
    }

    SDL_UnlockAudio();
}

void S_StopSound(int entnum, int entchannel)
{
    int i;

    SDL_LockAudio();

    for (i=0 ; i<MAX_DYNAMIC_CHANNELS ; i++)
    {
        if (channels[i].entnum == entnum
            && channels[i].entchannel == entchannel)
        {
            channels[i].sfx = NULL;
            channels[i].pos = 999999;
            SDL_UnlockAudio();
            return;
        }
    }

    SDL_UnlockAudio();
}

void S_StopAllSounds(void)
{
    int        i;

    if (!sound_started)
        return;

    SDL_LockAudio();

    total_channels = MAX_DYNAMIC_CHANNELS + NUM_AMBIENTS;    // no statics

    for (i=0 ; i<MAX_CHANNELS ; i++)
    {
        if (channels[i].sfx)
            channels[i].sfx = NULL;
        channels[i].pos = 999999;
    }

    memset(channels, 0, MAX_CHANNELS * sizeof(channel_t));

    SDL_UnlockAudio();
}



/*
=================
S_StaticSound
=================
*/
void S_StaticSound (int soundindex, vec3_t origin, float vol, float attenuation)
{
    sfx_t *sfx;
    channel_t *ss;
    sfxcache_t *sc;

    if (soundindex < 0 || soundindex >= num_sfx)
        return;

    sfx = known_sfx + soundindex;
    if (!sfx)
        return;

    if (total_channels == MAX_CHANNELS)
    {
        Con_Printf ("total_channels == MAX_CHANNELS\n");
        return;
    }

    SDL_LockAudio();

    ss = &channels[total_channels];
    total_channels++;

    sc = S_LoadSound (sfx);
    if (!sc)
    {
        SDL_UnlockAudio();
        return;
    }

    ss->pos = 0;
    ss->looping = 1;
    ss->sfx = sfx;
    VectorCopy (origin, ss->origin);
    ss->master_vol = vol;
    ss->dist_mult = (attenuation/64) / sound_nominal_clip_dist;

    S_Spatialize (ss);

    SDL_UnlockAudio();
}


//=============================================================================

/*
===================
S_UpdateAmbientSounds
===================
*/
void S_UpdateAmbientSounds (void)
{
    mleaf_t        *l;
    float        vol;
    int            ambient_channel;
    channel_t    *chan;

    // calc ambient sound levels
    if (!cl.worldmodel)
        return;

    l = Mod_PointInLeaf (listener_origin, cl.worldmodel);
    if (!l || !Cvar_VariableValue("ambient_level"))
    {
        for (ambient_channel = 0 ; ambient_channel< NUM_AMBIENTS ; ambient_channel++)
            channels[ambient_channel].sfx = NULL;
        return;
    }

    for (ambient_channel = 0 ; ambient_channel< NUM_AMBIENTS ; ambient_channel++)
    {
        chan = &channels[ambient_channel];
        chan->sfx = ambient_sfx[ambient_channel];
        chan->looping = 1;

        vol = Cvar_VariableValue("ambient_level") * l->ambient_sound_level[ambient_channel];
        if (vol < 8)
            vol = 0;

    // don't adjust volume too fast
        if (chan->master_vol < vol)
        {
            chan->master_vol += host_frametime * Cvar_VariableValue("ambient_fade");
            if (chan->master_vol > vol)
                chan->master_vol = vol;
        }
        else if (chan->master_vol > vol)
        {
            chan->master_vol -= host_frametime * Cvar_VariableValue("ambient_fade");
            if (chan->master_vol < vol)
                chan->master_vol = vol;
        }

        chan->leftvol = chan->rightvol = chan->master_vol;
    }
}


/*
============
S_Update

Called once each time through the main loop
============
*/
void S_Update(vec3_t origin, vec3_t forward, vec3_t right, vec3_t up)
{
    int            i, j;
    channel_t    *ch;
    channel_t    *combine;

    if (!sound_started)
        return;

    SDL_LockAudio();

    VectorCopy(origin, listener_origin);
    VectorCopy(forward, listener_forward);
    VectorCopy(right, listener_right);
    VectorCopy(up, listener_up);

    // update general area ambient sound sources
    S_UpdateAmbientSounds ();

    combine = NULL;

    // update spatialization for static and dynamic sounds
    ch = channels+NUM_AMBIENTS;
    for (i=NUM_AMBIENTS ; i<total_channels; i++, ch++)
    {
        if (!ch->sfx)
            continue;
        S_Spatialize(ch);         // respatialize channel
        if (!ch->leftvol && !ch->rightvol)
            continue;

        // try to combine static sounds with a previous channel of the same
        // sound effect so we don't mix five torches every frame
        if (i >= MAX_DYNAMIC_CHANNELS + NUM_AMBIENTS)
        {
            // see if it can just use the last one
            if (combine && combine->sfx == ch->sfx)
            {
                combine->leftvol += ch->leftvol;
                combine->rightvol += ch->rightvol;
                ch->leftvol = ch->rightvol = 0;
                continue;
            }
            // search for one
            combine = channels+MAX_DYNAMIC_CHANNELS + NUM_AMBIENTS;
            for (j=MAX_DYNAMIC_CHANNELS + NUM_AMBIENTS ; j<i; j++, combine++)
                if (combine->sfx == ch->sfx)
                    break;

            if (j == total_channels)
            {
                combine = NULL;
            }
            else
            {
                if (combine != ch)
                {
                    combine->leftvol += ch->leftvol;
                    combine->rightvol += ch->rightvol;
                    ch->leftvol = ch->rightvol = 0;
                }
                continue;
            }
        }
    }

    SDL_UnlockAudio();
}

/*
===============================================================================

console functions

===============================================================================
*/

void S_Play(void)
{
    static int hash=345;
    std::string name;
    int sfx;
    
    size_t i = 1;
    while (i < Cmd_Argc())
    {
        name = Filesystem::DefaultPathExtension(Cmd_Argv(i), ".wav");
        sfx = S_PrecacheSound(name);
        S_StartSound(hash++, 0, sfx, listener_origin, 1.0, 1.0);
        i++;
    }
}

void S_PlayVol(void)
{
    static int hash=543;
    float vol;
    std::string name;
    int sfx;

    size_t i = 1;
    while (i < Cmd_Argc())
    {
        name = Filesystem::DefaultPathExtension(Cmd_Argv(i), ".wav");
        sfx = S_PrecacheSound(name);
        vol = Util::lexical_cast<float>(Cmd_Argv(i+1));
        S_StartSound(hash++, 0, sfx, listener_origin, vol, 1.0);
        i += 2;
    }
}



void S_LocalSound(const std::string &sound)
{
    int sfx;

    if (Cvar_VariableValue("nosound"))
        return;
    if (!sound_started)
        return;

    sfx = S_PrecacheSound (sound);
    if (!sfx)
    {
        Con_Printf ("S_LocalSound: can't cache %s\n", sound.c_str());
        return;
    }
    S_StartSound (cl.viewentity, -1, sfx, vec3_origin, 1, 1);
}
